!
! *
! *  Elmer, A Finite Element Software for Multiphysical Problems
! *
! *  Copyright 1st April 1995 - , CSC - IT Center for Science Ltd., Finland
! * 
! *  This program is free software; you can redistribute it and/or
! *  modify it under the terms of the GNU General Public License
! *  as published by the Free Software Foundation; either version 2
! *  of the License, or (at your option) any later version.
! * 
! *  This program is distributed in the hope that it will be useful,
! *  but WITHOUT ANY WARRANTY; without even the implied warranty of
! *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! *  GNU General Public License for more details.
! *
! *  You should have received a copy of the GNU General Public License
! *  along with this program (in file fem/GPL-2); if not, write to the 
! *  Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
! *  Boston, MA 02110-1301, USA.
! *
! *****************************************************************************/
!
!

!------------------------------------------------------------------------------
   SUBROUTINE ElmerSolver(initialize)
!------------------------------------------------------------------------------

     USE MainUtils

!------------------------------------------------------------------------------
     IMPLICIT NONE
!------------------------------------------------------------------------------

     INTEGER :: Initialize

!------------------------------------------------------------------------------
!    Local variables
!------------------------------------------------------------------------------

     INTEGER :: i,j,k,n,l,t,k1,k2,iter,Ndeg,istat,nproc

     REAL(KIND=dp) :: s,dt,dtfunc
     REAL(KIND=dP), POINTER :: WorkA(:,:,:) => NULL()
     REAL(KIND=dp), POINTER, SAVE :: sTime(:), sStep(:), sInterval(:), sSize(:), &
                               steadyIt(:),nonlinIt(:),sPrevSizes(:,:)

     TYPE(Element_t),POINTER :: CurrentElement

     LOGICAL :: GotIt,Transient,Scanning,LastSaved

     INTEGER :: TimeIntervals,interval,timestep, &
       TotalTimesteps,SavedSteps,CoupledMaxIter,CoupledMinIter

     INTEGER, POINTER, SAVE :: Timesteps(:),OutputIntervals(:), &
                 OutputMask(:),ActiveSolvers(:)
     REAL(KIND=dp), POINTER, SAVE :: TimestepSizes(:,:)

     INTEGER(KIND=AddrInt) :: ControlProcedure

     LOGICAL :: InitDirichlet, ExecThis

     TYPE(ElementType_t),POINTER :: elmt

     TYPE(ParEnv_t), POINTER :: ParallelEnv

     CHARACTER(LEN=MAX_NAME_LEN) :: ModelName, eq, ExecCommand
     CHARACTER(LEN=MAX_PATH_LEN) :: OutputFile, PostFile, RestartFile, &
                OutputName=' ',PostName=' ', When, OptionString

     TYPE(Variable_t), POINTER :: Var
     TYPE(Mesh_t), POINTER :: Mesh
     TYPE(Solver_t), POINTER :: Solver

     REAL(KIND=dp) :: RealTime,tt

     LOGICAL :: FirstLoad = .TRUE., FirstTime=.TRUE.

     INTEGER :: iargc

     !
     ! If parallel execution requested, initialize parallel environment:
     !------------------------------------------------------------------
     ParallelEnv => ParallelInit()
     OutputPE = ParEnv % MyPE

!tt = realtime()
     IF ( FirstTime ) THEN
       !
       ! Print banner to output:




























































































































































































       ! -----------------------
       CALL Info( 'MAIN', ' ', Level=3 )
       CALL Info( 'MAIN', '==========================================', Level=3 )
       CALL Info( 'MAIN', ' E L M E R  S O L V E R  S T A R T I N G',  Level=3  )
       CALL Info( 'MAIN', ' Library version: ' // "6.1" &
                    // ' (Rev: ' // "" // ')' , Level=3 )
       IF ( ParEnv % PEs > 1 ) &
         CALL Info( 'MAIN', ' Running in parallel using ' // &
          TRIM(i2s(ParEnv % PEs)) // ' tasks.', Level=3 )
       CALL Info( 'MAIN', ' HYPRE library linked in.', Level=3 )
       CALL Info( 'MAIN', '==========================================', Level=3 )

       IF ( IARGC() > 0 ) THEN
         CALL getarg( 1,OptionString )
         IF ( OptionString=='-v' .OR. OptionString == '--version' ) RETURN
       END IF

       CALL InitializeElementDescriptions
       FirstTime = .FALSE.
     END IF

     !
     ! Read input file name:
     !----------------------
     IF ( ParEnv % PEs <= 1 ) THEN
       IF ( IARGC() > 0 ) THEN
         CALL getarg( 1,ModelName )
         IF ( IARGC() > 1 ) CALL getarg( 2,eq )
       ELSE
         OPEN( 1, File='ELMERSOLVER_STARTINFO', STATUS='OLD', ERR=10 )
         READ(1,'(a)') ModelName
         CLOSE(1)
       END IF
     ELSE
       OPEN( 1, File='ELMERSOLVER_STARTINFO', STATUS='OLD', ERR=10 )
       READ(1,'(a)') ModelName
       CLOSE(1)
     END IF

!------------------------------------------------------------------------------
!    Read element definition file, and initialize element types
!------------------------------------------------------------------------------
     IF ( Initialize==1 ) THEN
       CALL FreeModel(CurrentModel)
       FirstLoad=.TRUE.
     END IF

!------------------------------------------------------------------------------
!    Read Model and mesh from Elmer mesh data base
!------------------------------------------------------------------------------
     DO WHILE( .TRUE. )

       IF ( initialize==2 ) GOTO 1

       IF ( FirstLoad ) THEN
          CALL Info( 'MAIN', ' ', Level = 3 )
          CALL Info( 'MAIN', ' ', Level = 3 )
          CALL Info( 'MAIN', '-----------------------', Level = 3 )
          CALL Info( 'MAIN', 'Reading Model ... ', Level = 3 )

          INQUIRE(Unit=InFileUnit, Opened=GotIt)
          IF ( gotIt ) CLOSE(inFileUnit)

          OPEN( Unit=InFileUnit, Action='Read',File=ModelName,Status='OLD',ERR=20 )
          CurrentModel => LoadModel( ModelName,.FALSE.,ParEnv % PEs,ParEnv % MyPE )

          CALL Info( 'MAIN', 'Done               ', Level = 3 )
          CALL Info( 'MAIN', '-----------------------', Level = 3 )
       ELSE
         IF ( Initialize==3 ) THEN
           INQUIRE(Unit=InFileUnit, Opened=GotIt)
           IF ( gotIt ) CLOSE(inFileUnit)
           OPEN( Unit=InFileUnit, Action='Read', & 
                    File=ModelName,Status='OLD',ERR=20 )
         END IF

         IF ( .NOT.ReloadInputFile(CurrentModel) ) EXIT

         Mesh => CurrentModel % Meshes
         DO WHILE( ASSOCIATED(Mesh) )
           Mesh % SavesDone = 0
           Mesh => Mesh % Next
         END DO
       END IF

1      CONTINUE

       CALL ListAddLogical( CurrentModel % Simulation, &
             'Initialization Phase', .TRUE. )

!------------------------------------------------------------------------------
!      Check for transient case
!------------------------------------------------------------------------------
       eq = ListGetString( CurrentModel % Simulation, 'Simulation Type', GotIt )
       Scanning  = eq == 'scanning'
       Transient = eq == 'transient'

!------------------------------------------------------------------------------
!      Initialize the log file output system
!------------------------------------------------------------------------------
       MinOutputLevel = ListGetInteger( CurrentModel % Simulation, &
                  'Min Output Level', GotIt )

       MaxOutputLevel = ListGetInteger( CurrentModel % Simulation, &
                 'Max Output Level', GotIt )

       IF ( .NOT. GotIt ) MaxOutputLevel = 32

       OutputMask => ListGetIntegerArray( CurrentModel % Simulation, &
                    'Output Level', GotIt )

       IF ( GotIt ) THEN
          DO i=1,SIZE(OutputMask)
             OutputLevelMask(i-1) = OutputMask(i) /= 0
          END DO
       END IF

       DO i=0,31
          OutputLevelMask(i) = OutputLevelMask(i) .AND. &
           i >= MinOutputLevel .AND. i <= MaxOutputLevel
       END DO

       OutputPrefix = ListGetLogical( CurrentModel % Simulation, &
                      'Output Prefix', GotIt )
       IF ( .NOT. GotIt ) OutputPrefix = .FALSE.

       OutputCaller = ListGetLogical( CurrentModel % Simulation, &
                      'Output Caller', GotIt )
       IF ( .NOT. GotIt ) OutputCaller = .TRUE.
!------------------------------------------------------------------------------

!------------------------------------------------------------------------------
!      Figure out what (flow,heat,stress,...) should be computed, and get
!      memory for the dofs
!------------------------------------------------------------------------------
       CALL AddSolvers

!------------------------------------------------------------------------------
!      Time integration and/or steady state steps
!------------------------------------------------------------------------------
       IF ( Transient .OR. Scanning ) THEN
         Timesteps => ListGetIntegerArray( CurrentModel % Simulation, &
                       'Timestep Intervals', GotIt )

         IF ( .NOT.GotIt ) THEN
           CALL Fatal( ' ', 'Keyword [Timestep Intervals] MUST be ' //  &
                   'defined for time dependent simulations' )
         END IF 

         TimestepSizes => ListGetConstRealArray( CurrentModel % Simulation, &
                               'Timestep Sizes', GotIt )
         IF ( .NOT.GotIt ) THEN
           IF( Scanning ) THEN
             ALLOCATE(TimestepSizes(1,1))
             TimestepSizes(1,1) = 1.0_dp
           ELSE
             CALL Fatal( ' ', 'Keyword [Timestep Sizes] MUST be ' //  &
                 'defined for time dependent simulations' )
             STOP
           END IF
         END IF 

         TimeIntervals = SIZE(Timesteps)

         CoupledMaxIter = ListGetInteger( CurrentModel % Simulation, &
               'Steady State Max Iterations', GotIt, minv=1 )
         IF ( .NOT. GotIt ) CoupledMaxIter = 1
!------------------------------------------------------------------------------
       ELSE
!------------------------------------------------------------------------------
!        Steady state
!------------------------------------------------------------------------------
         ALLOCATE(Timesteps(1))

         Timesteps(1) = ListGetInteger( CurrentModel % Simulation, &
               'Steady State Max Iterations', GotIt,minv=1 )
         IF ( .NOT. GotIt ) Timesteps(1)=1
  
         ALLOCATE(TimestepSizes(1,1))
         TimestepSizes(1,1) = 1.0D0

         CoupledMaxIter = 1
         TimeIntervals  = 1
       END IF

       IF ( FirstLoad ) &
         ALLOCATE( sTime(1), sStep(1), sInterval(1), sSize(1), &
             steadyIt(1), nonLinit(1), sPrevSizes(1,5) )

       dt   = 0._dp

       sTime = 0._dp
       sStep = 0

       sSize = dt
       sPrevSizes = 0_dp

       sInterval = 0._dp

       steadyIt = 0
       nonlinIt = 0

       CoupledMinIter = ListGetInteger( CurrentModel % Simulation, &
                  'Steady State Min Iterations', GotIt )

!------------------------------------------------------------------------------
!      Add coordinates and simulation time to list of variables so that
!      coordinate dependent parameter computing routines can ask for
!      them...
!------------------------------------------------------------------------------
       IF ( FirstLoad ) CALL AddMeshCoordinatesAndTime

!------------------------------------------------------------------------------
!      Get Output File Options
!------------------------------------------------------------------------------

       OutputIntervals => ListGetIntegerArray( CurrentModel % Simulation, &
                       'Output Intervals', GotIt )
       IF ( .NOT. GotIt ) THEN
         ALLOCATE( OutputIntervals(SIZE(TimeSteps)) )
         OutputIntervals = 1
       END IF


       ! Initial Conditions:
       ! -------------------
       IF ( FirstLoad ) CALL SetInitialConditions

       TotalTimesteps = 0
       LastSaved = .TRUE.
       DO interval=1,TimeIntervals
         DO timestep = 1,Timesteps(interval)
           IF( OutputIntervals(Interval) == 0 ) CYCLE
           LastSaved = .FALSE.
           IF ( MOD(Timestep-1, OutputIntervals(Interval))==0 ) THEN
              LastSaved = .TRUE.
              TotalTimesteps = TotalTimesteps + 1
           END IF
         END DO
       END DO

       DO i=1,CurrentModel % NumberOfSolvers
          Solver => CurrentModel % Solvers(i)
          When = ListGetString( Solver % Values, 'Exec Solver', GotIt )
          IF ( GotIt ) THEN
             IF ( When == 'after simulation' .OR. When == 'after all' ) THEN
                LastSaved = .FALSE.
             END IF
          ELSE
           IF ( Solver % SolverExecWhen == SOLVER_EXEC_AFTER_ALL ) THEN
              LastSaved = .FALSE.
           END IF
          END IF
       END DO

       IF ( .NOT.LastSaved ) TotalTimesteps = TotalTimesteps + 1
       IF( TotalTimesteps == 0 ) TotalTimesteps = 1

       CALL ListAddLogical( CurrentModel % Simulation,  &
            'Initialization Phase', .FALSE. )

       FirstLoad = .FALSE.
       IF ( Initialize==1 ) EXIT

!if ( parenv % mype==0 ) print*,'INIT TIME: ', realtime()-tt; tt = realtime()
!------------------------------------------------------------------------------
!      Here we actually start the simulation ....
!      First go trough timeintervals
!------------------------------------------------------------------------------
       ExecCommand = ListGetString( CurrentModel % Simulation, &
                 'Control Procedure', GotIt )
       IF ( GotIt ) THEN
          ControlProcedure = GetProcAddr( ExecCommand )
          CALL ExecSimulationProc( ControlProcedure, CurrentModel )
       ELSE
          CALL ExecSimulation( TimeIntervals, CoupledMinIter, &
              CoupledMaxIter, OutputIntervals, Transient, Scanning)
       END IF
!if ( parenv % mype == 0 ) print*,'SOLVE TIME: ', realtime()-tt
!------------------------------------------------------------------------------
!    Always save the last step to output
!------------------------------------------------------------------------------
       IF ( .NOT.LastSaved ) THEN
         DO i=1,CurrentModel % NumberOfSolvers
            Solver => CurrentModel % Solvers(i)
            IF ( Solver % Procedure == 0 ) CYCLE
            ExecThis = ( Solver % SolverExecWhen == SOLVER_EXEC_AHEAD_SAVE)
            When = ListGetString( Solver % Values, 'Exec Solver', GotIt )
            IF ( GotIt ) ExecThis = ( When == 'before saving') 
            IF( ExecThis ) CALL SolverActivate( CurrentModel,Solver,dt,Transient )
         END DO

         CALL SaveToPost(0)
         CALL SaveCurrent(Timestep)

         DO i=1,CurrentModel % NumberOfSolvers
            Solver => CurrentModel % Solvers(i)
            IF ( Solver % Procedure == 0 ) CYCLE
            ExecThis = ( Solver % SolverExecWhen == SOLVER_EXEC_AFTER_SAVE)
            When = ListGetString( Solver % Values, 'Exec Solver', GotIt )
            IF ( GotIt ) ExecThis = ( When == 'after saving') 
            IF( ExecThis ) CALL SolverActivate( CurrentModel,Solver,dt,Transient )
         END DO
       END IF

       IF ( Initialize>=2 ) EXIT
     END DO
!------------------------------------------------------------------------------
!    THIS IS THE END (...,at last, the end, my friend,...)
!------------------------------------------------------------------------------
     IF ( Initialize /= 1 ) CALL Info( 'ElmerSolver', '*** Elmer Solver: ALL DONE ***' )
     IF ( ParEnv % PEs>1 )  CALL ParallelFinalize()

     IF ( Initialize <=0 ) CALL FreeModel(CurrentModel)

     CALL Info('ElmerSolver','The end')

     RETURN

10   CONTINUE
     CALL Fatal( 'ElmerSolver', 'Unable to find ELMERSOLVER_STARTINFO, can not execute.' )

20   CONTINUE
     CALL Fatal( 'ElmerSolver', 'Unable to find input file [' // &
              TRIM(Modelname) // '], can not execute.' )

   CONTAINS 

!------------------------------------------------------------------------------
    SUBROUTINE AddSolvers
!------------------------------------------------------------------------------
      INTEGER :: i,j,k,nlen
      LOGICAL :: InitSolver, Found
!------------------------------------------------------------------------------
      DO i=1,CurrentModel % NumberOfSolvers
        eq = ListGetString( CurrentModel % Solvers(i) % Values,'Equation', Found )
        nlen = LEN_TRIM(eq)
     
       IF ( Found ) THEN
          DO j=1,CurrentModel % NumberOFEquations
             ActiveSolvers => ListGetIntegerArray( CurrentModel % Equations(j) % Values, &
                                'Active Solvers', Found )
             IF ( Found ) THEN
                DO k=1,SIZE(ActiveSolvers)
                   IF ( ActiveSolvers(k) == i ) THEN
                      CALL ListAddLogical( CurrentModel % Equations(j) % Values, eq(1:nlen), .TRUE. )
                      EXIT
                   END IF
                END DO
             END IF
          END DO
       END IF
     END DO

     DO i=1,CurrentModel % NumberOfSolvers
        eq = ListGetString( CurrentModel % Solvers(i) % Values,'Equation', Found )

        Solver => CurrentModel % Solvers(i)
        InitSolver = ListGetLogical( Solver % Values, 'Initialize', Found )
        IF ( Found .AND. InitSolver ) THEN
          CALL FreeMatrix( Solver % Matrix )
          CALL ListAddLogical( Solver % Values, 'Initialize', .FALSE. )
        END IF

        IF ( Solver % Procedure == 0 .OR. InitSolver ) THEN
          IF ( .NOT. ASSOCIATED( Solver % Mesh ) ) THEN
            Solver % Mesh => CurrentModel % Meshes
          END IF
          CurrentModel % Solver => Solver
          CALL AddEquationBasics( Solver, eq, Transient )
          CALL AddEquationSolution( Solver, Transient )
        END IF
     END DO
!------------------------------------------------------------------------------
  END SUBROUTINE AddSolvers
!------------------------------------------------------------------------------


!------------------------------------------------------------------------------
  SUBROUTINE AddMeshCoordinatesAndTime
!------------------------------------------------------------------------------
     TYPE(Variable_t), POINTER :: DtVar

     NULLIFY( Solver )

     Mesh => CurrentModel % Meshes 
     DO WHILE( ASSOCIATED( Mesh ) )
       CALL VariableAdd( Mesh % Variables, Mesh,Solver, &
             'Coordinate 1',1,Mesh % Nodes % x )

       CALL VariableAdd(Mesh % Variables,Mesh,Solver, &
             'Coordinate 2',1,Mesh % Nodes % y )

       CALL VariableAdd(Mesh % Variables,Mesh,Solver, &
             'Coordinate 3',1,Mesh % Nodes % z )

       CALL VariableAdd( Mesh % Variables, Mesh, Solver, 'Time', 1, sTime )
       CALL VariableAdd( Mesh % Variables, Mesh, Solver, 'Timestep', 1, sStep )
       CALL VariableAdd( Mesh % Variables, Mesh, Solver, 'Timestep size', 1, sSize )
       CALL VariableAdd( Mesh % Variables, Mesh, Solver, 'Timestep interval', 1, sInterval )

       ! Save some previous timesteps for variable timestep multistep methods
       DtVar => VariableGet( Mesh % Variables, 'Timestep size' )
       DtVar % PrevValues => sPrevSizes

       CALL VariableAdd( Mesh % Variables, Mesh, Solver, &
               'nonlin iter', 1, nonlinIt )
       CALL VariableAdd( Mesh % Variables, Mesh, Solver, &
               'coupled iter', 1, steadyIt )
       Mesh => Mesh % Next
     END DO
!------------------------------------------------------------------------------
  END SUBROUTINE AddMeshCoordinatesAndTime
!------------------------------------------------------------------------------


!------------------------------------------------------------------------------
   SUBROUTINE SetInitialConditions
!------------------------------------------------------------------------------
     USE DefUtils
     INTEGER :: DOFs
     CHARACTER(LEN=MAX_NAME_LEN) :: str
     LOGICAL :: Found
     TYPE(Solver_t), POINTER :: Solver
     INTEGER, ALLOCATABLE :: Indexes(:)
     REAL(KIND=dp),ALLOCATABLE :: Work(:)

     INTEGER :: i,j,k,l,m,vect_dof,real_dof,dim

     REAL(KIND=dp) :: nrm(3),t1(3),t2(3),vec(3),tmp(3),udot
     TYPE(ValueList_t), POINTER :: BC
     TYPE(Nodes_t), SAVE :: Nodes
     LOGICAL :: nt_boundary
     TYPE(Element_t), POINTER :: Element
     TYPE(Variable_t), POINTER :: var, vect_var

     dim = CoordinateSystemDimension()

     IF (GetLogical(GetSimulation(),'Restart Before Initial Conditions',Found)) THEN
       CALL Restart
       CALL InitCond
     ELSE
       CALL InitCond
       CALL Restart
     END IF

!------------------------------------------------------------------------------
!    Make sure that initial values at boundaries are set correctly.
!    NOTE: This overrides the initial condition setting for field variables!!!!
!-------------------------------------------------------------------------------
     InitDirichlet = ListGetLogical( CurrentModel % Simulation, &
            'Initialize Dirichlet Conditions', GotIt ) 
     IF ( .NOT. GotIt ) InitDirichlet = .TRUE.

     vect_var => Null()
     IF ( InitDirichlet ) THEN
       Mesh => CurrentModel % Meshes
       DO WHILE( ASSOCIATED(Mesh) )
         ALLOCATE( Work(Mesh % MaxElementDOFs) )
         CALL SetCurrentMesh( CurrentModel, Mesh )

         DO t = Mesh % NumberOfBulkElements + 1, &
                 Mesh % NumberOfBulkElements + Mesh % NumberOfBoundaryElements

           Element => Mesh % Elements(t)

           ! Set also the current element pointer in the model structure to
           ! reflect the element being processed:
           ! ---------------------------------------------------------------
           CurrentModel % CurrentElement => Element
           n = Element % TYPE % NumberOfNodes

           BC => GetBC()

           Var => Mesh % Variables
           DO WHILE( ASSOCIATED(Var) )
             Solver => Var % Solver
             IF ( .NOT. ASSOCIATED(Solver) ) Solver => CurrentModel % Solver

             str = ListGetString( Solver % Values, 'Namespace', Found )
             IF (Found) CALL ListSetNamespace(TRIM(str))


             IF ( Var % DOFs <= 1 ) THEN
               Work(1:n) = GetReal( BC,Var % Name, gotIt )
               IF ( GotIt ) THEN

                 nt_boundary = .FALSE.
                 IF ( GetElementFamily() /= 1 ) THEN
                   k = LEN_TRIM(var % name)
                   vect_dof = ICHAR(Var % Name(k:k))-ICHAR('0');
                   IF ( vect_dof>=1 .AND. vect_dof<= 3 ) THEN
                     nt_boundary =  GetLogical( BC, &
                        'normal-tangential '//var % name(1:k-2), gotIt)

                     IF ( nt_boundary ) THEN
                       nt_boundary = .FALSE.
                       Vect_var => Mesh % Variables
                       DO WHILE( ASSOCIATED(Vect_var) )
                         IF ( Vect_var % Dofs>=dim ) THEN
                           DO real_dof=1,Vect_var % Dofs
                             nt_boundary = ASSOCIATED(Var % Values, &
                               Vect_var % Values(real_dof::Vect_var % Dofs))
                             IF ( nt_boundary ) EXIT
                           END DO
                           IF ( nt_boundary ) EXIT
                         END IF
                         Vect_var => Vect_var % Next
                       END DO
                     END IF

                     IF ( nt_boundary ) THEN
                       CALL GetElementNodes(Nodes)
                       nrm = NormalVector(Element,Nodes,0._dp,0._dp,.TRUE.)
                       SELECT CASE(Element % Type % Dimension)
                       CASE(1)
                         t1(1) =  nrm(2)
                         t1(2) = -nrm(1)
                       CASE(2)
                         CALL TangentDirections(nrm,t1,t2)
                       END SELECT

                       SELECT CASE(vect_dof)
                         CASE(1)
                           vec = nrm
                         CASE(2)
                           vec = t1
                         CASE(3)
                           vec = t2
                       END SELECT
                     END IF
                   END IF
                 END IF

                 DO j=1,n
                   k = Element % NodeIndexes(j)
                   IF ( ASSOCIATED(Var % Perm) ) k = Var % Perm(k)
                   IF ( k>0 ) THEN
                     IF ( nt_boundary ) THEN
                       DO l=1,dim
                         m = l+real_dof-vect_dof
                         tmp(l)=Vect_var % Values(Vect_var % Dofs*(k-1)+m)
                       END DO
                       udot = SUM(vec(1:dim)*tmp(1:dim))
                       tmp(1:dim)=tmp(1:dim)+(work(j)-udot)*vec(1:dim)
                       DO l=1,dim
                         m = l+real_dof-vect_dof
                         Vect_var % Values(Vect_var % Dofs*(k-1)+m)=tmp(l)
                       END DO
                     ELSE
                       Var % Values(k) = Work(j)
                     END IF
                   END IF
                 END DO
               END IF

               IF ( Transient .AND. Solver % TimeOrder==2 ) THEN
                  Work(1:n) = GetReal( BC, TRIM(Var % Name) // ' Velocity', GotIt )
                  IF ( GotIt ) THEN
                    DO j=1,n
                      k = Element % NodeIndexes(j)
                      IF ( ASSOCIATED(Var % Perm) ) k = Var % Perm(k)
                      IF ( k>0 ) Var % PrevValues(k,1) = Work(j)
                    END DO
                  END IF
                  Work(1:n) = GetReal( BC, TRIM(Var % Name) // ' Acceleration', GotIt )
                  IF ( GotIt ) THEN
                    DO j=1,n
                      k = Element % NodeIndexes(j)
                      IF ( ASSOCIATED(Var % Perm) ) k = Var % Perm(k)
                      IF ( k>0 ) Var % PrevValues(k,2) = Work(j)
                    END DO
                  END IF
               END IF
             ELSE
               CALL ListGetRealArray( BC, &
                 Var % Name, WorkA, n, Element % NodeIndexes, gotIt )
               IF ( GotIt ) THEN
                 DO j=1,n
                   k = Element % NodeIndexes(j)
                   DO l=1,MIN(SIZE(WorkA,1),Var % DOFs)
                     IF ( ASSOCIATED(Var % Perm) ) k = Var % Perm(k)
                     IF ( k>0 ) Var % Values(Var % DOFs*(k-1)+l) = WorkA(l,1,j)
                   END DO
                 END DO
               ELSE
               END IF
             END IF

             CALL ListSetNamespace('')
             Var => Var % Next
           END DO
         END DO
         DEALLOCATE( Work )
         Mesh => Mesh % Next
       END DO
     END IF
!------------------------------------------------------------------------------
   END SUBROUTINE SetInitialConditions
!------------------------------------------------------------------------------


!------------------------------------------------------------------------------
   SUBROUTINE InitCond
!------------------------------------------------------------------------------
     USE DefUtils
     TYPE(Element_t), POINTER :: Edge
     INTEGER :: DOFs,i,j,k,l
     CHARACTER(LEN=MAX_NAME_LEN) :: str
     LOGICAL :: Found
     TYPE(Solver_t), POINTER :: Solver
     INTEGER, ALLOCATABLE :: Indexes(:)
     REAL(KIND=dp),ALLOCATABLE :: Work(:)
!------------------------------------------------------------------------------
     DO i=1,CurrentModel % NumberOfBodies
       Mesh => CurrentModel % Meshes
       DO WHILE( ASSOCIATED( Mesh ) )
         ALLOCATE( Indexes(Mesh % MaxElementDOFs), Work(Mesh % MaxElementDOFs) )
         CALL SetCurrentMesh( CurrentModel, Mesh )

         DO t=1, Mesh % NumberOfBulkElements+Mesh % NumberOfBoundaryElements
           CurrentElement =>  Mesh % Elements(t)
           CurrentModel % CurrentElement => CurrentElement

           IF ( CurrentElement % BodyId == i ) THEN
             j = ListGetInteger(CurrentModel % Bodies(i) % Values, &
                'Initial Condition',GotIt, 1, CurrentModel % NumberOfICs )

             IF ( GotIt ) THEN
               n    = GetElementNOFNodes()

               Var => Mesh % Variables
               DO WHILE( ASSOCIATED(Var) ) 
                 IF ( ASSOCIATED( Var % Solver ) ) THEN
                   DOFs = GetElementDOFs( Indexes, USolver=Var % Solver )
                 ELSE
                   DOFs = GetElementDOFs( Indexes )
                 END IF

                 Solver => Var % Solver
                 IF ( .NOT. ASSOCIATED(Solver) ) Solver => CurrentModel % Solver

                 str = ListGetString( Solver % Values, 'Namespace', Found )
                 IF (Found) CALL ListSetNamespace(TRIM(str))


! Note: this is for global variables but it is not fully generic yet 
                 IF( SIZE( Var % Values ) == Var % DOFs ) THEN

                    Var % Values = ListGetCReal( CurrentModel % ICs(j) % Values, &
                             Var % Name, GotIt )


                 ELSE IF ( Var % DOFs <= 1 ) THEN
                    Work(1:n) = GetReal( CurrentModel % ICs(j) % Values, &
                           Var % Name, GotIt )
                    IF ( GotIt ) THEN
                      DO k=1,n
                        k1 = Indexes(k)
                        IF ( ASSOCIATED(Var % Perm) ) k1 = Var % Perm(k1)
                        IF ( k1>0 ) Var % Values(k1) = Work(k)
                      END DO
                    END IF

                    IF ( Transient .AND. Solver % TimeOrder==2 ) THEN
                      Work(1:n) = GetReal( CurrentModel % ICs(j) % Values, &
                             TRIM(Var % Name) // ' Velocity', GotIt )
                      IF ( GotIt ) THEN
                        DO k=1,n
                          k1 = Indexes(k)
                          IF ( ASSOCIATED(Var % Perm) ) k1 = Var % Perm(k1)
                          IF ( k1>0 ) Var % PrevValues(k1,1) = Work(k)
                        END DO
                      END IF
                      Work(1:n) = GetReal( CurrentModel % ICs(j) % Values, &
                             TRIM(Var % Name) // ' Acceleration', GotIt )
                      IF ( GotIt ) THEN
                        DO k=1,n
                          k1 = Indexes(k)
                          IF ( ASSOCIATED(Var % Perm) ) k1 = Var % Perm(k1)
                          IF ( k1>0 ) Var % PrevValues(k1,2) = Work(k)
                        END DO
                      END IF
                    END IF

                    IF(ASSOCIATED(Mesh % Edges)) THEN
                      IF ( i<=Mesh % NumberOfBulkElements) THEN
                        Gotit = ListCheckPresent( CurrentModel % ICs(j) % Values, &
                                  TRIM(Var % Name)//' {e}' )
                        IF ( Gotit ) THEN
                          DO k=1,CurrentElement % Type % NumberOfedges
                            Edge => Mesh % Edges(CurrentElement % EdgeIndexes(k))
                            l = Var % Perm(CurrentElement % EdgeIndexes(k)+Mesh % NumberOfNodes)
                            IF ( l>0 ) THEN
                              CALL LocalBcIntegral( CurrentModel % ICs(j) % Values, &
                                  Edge, Edge % Type % NumberOfNodes, CurrentElement, n, &
                                      TRIM(Var % Name)//' {e}', Work(1) )
                              Var % Values(l) = Work(1)
                            END IF
                          END DO
                        END IF
                      END IF
                    END IF

                 ELSE
                    CALL ListGetRealArray( CurrentModel % ICs(j) % Values, &
                      Var % Name, WorkA, n, CurrentElement % NodeIndexes, gotIt )

                    IF ( GotIt ) THEN
                      DO k=1,n
                        k1 = Indexes(k)
                        DO l=1,MIN(SIZE(WorkA,1),Var % DOFs)
                          IF ( ASSOCIATED(Var % Perm) ) k1 = Var % Perm(k1)
                          IF ( k1>0 ) Var % Values(Var % DOFs*(k1-1)+l) = WorkA(l,1,k)
                        END DO
                      END DO
                    END IF
                 END IF
                 CALL ListSetNamespace('')
                 Var => Var % Next
               END DO
             END IF
           END IF
         END DO
         DEALLOCATE( Indexes, Work )
         Mesh => Mesh % Next
       END DO
     END DO
!------------------------------------------------------------------------------
   END SUBROUTINE InitCond
!------------------------------------------------------------------------------

!------------------------------------------------------------------------------
!    Check if we are restarting
!------------------------------------------------------------------------------
   SUBROUTINE Restart
!------------------------------------------------------------------------------
     USE DefUtils
     LOGICAL :: Gotit
     INTEGER :: k,StartTime
!------------------------------------------------------------------------------


        RestartFile = ListGetString( CurrentModel % Simulation, &
                 'Restart File', GotIt )

     IF ( GotIt ) THEN
       k = ListGetInteger( CurrentModel % Simulation,'Restart Position',GotIt, &
                  minv=0 )

       Mesh => CurrentModel % Meshes
       DO WHILE( ASSOCIATED(Mesh) ) 

         IF ( LEN_TRIM(Mesh % Name) > 0 ) THEN
           OutputName = TRIM(Mesh % Name) // '/' // TRIM(RestartFile)
         ELSE
           OutputName = TRIM(RestartFile)
         END IF
         IF ( ParEnv % PEs > 1 ) &
           OutputName = TRIM(OutputName) // '.' // TRIM(i2s(ParEnv % MyPe))

         CALL SetCurrentMesh( CurrentModel, Mesh )
         CALL LoadRestartFile( OutputName,k,Mesh )

         StartTime = ListGetConstReal( CurrentModel % Simulation,'Restart Time',GotIt)
         IF( GotIt ) THEN
	   Var  => VariableGet( Mesh % Variables, 'Time' )
           IF ( ASSOCIATED( Var ) )  Var % Values(1)  = StartTime
         END IF

         Mesh => Mesh % Next

       END DO


     END IF
!------------------------------------------------------------------------------
   END SUBROUTINE Restart
!------------------------------------------------------------------------------


!------------------------------------------------------------------------------
   SUBROUTINE ExecSimulation(TimeIntervals,  CoupledMinIter, &
              CoupledMaxIter, OutputIntervals, Transient, Scanning)
      INTEGER :: TimeIntervals,CoupledMinIter, CoupledMaxIter,OutputIntervals(:)
      LOGICAL :: Transient,Scanning
!------------------------------------------------------------------------------
     INTEGER :: interval, timestep, i, j, k, n
     REAL(KIND=dp) :: dt, ddt, dtfunc
     INTEGER :: timeleft,cum_timestep
     INTEGER, SAVE ::  stepcount=0, RealTimestep
     LOGICAL :: ExecThis,SteadyStateReached=.FALSE.

     REAL(KIND=dp) :: CumTime, MaxErr, AdaptiveLimit, &
           AdaptiveMinTimestep, AdaptiveMaxTimestep
     INTEGER :: SmallestCount, AdaptiveKeepSmallest, StepControl=-1
     LOGICAL :: AdaptiveTime = .TRUE.

     TYPE(Solver_t), POINTER :: Solver

     REAL(KIND=dp) :: RealTime, newtime, prevtime=0, maxtime
     REAL(KIND=dp), ALLOCATABLE :: xx(:,:), xxnrm(:), yynrm(:), PrevXX(:,:,:)

     DO i=1,CurrentModel % NumberOfSolvers
        Solver => CurrentModel % Solvers(i)
        IF ( Solver % Procedure==0 ) CYCLE
        IF ( Solver % SolverExecWhen == SOLVER_EXEC_AHEAD_ALL ) THEN
           CALL SolverActivate( CurrentModel,Solver,dt,Transient )
        END IF
     END DO

     DO interval = 1, TimeIntervals
        stepcount = stepcount + Timesteps(interval)
     END DO 

     cum_Timestep = 0
     ddt = 0.0d0
     DO interval = 1,TimeIntervals

!------------------------------------------------------------------------------
       IF ( Transient .OR. Scanning ) THEN
         dt = TimestepSizes(interval,1)
       ELSE
         dt = 1
       END IF
!------------------------------------------------------------------------------
!      go trough number of timesteps within an interval
!------------------------------------------------------------------------------
       RealTimestep = 1
       DO timestep = 1,Timesteps(interval)

         dtfunc = ListGetConstReal( CurrentModel % Simulation, &
                  'Timestep Function', gotIt)
         IF(GotIt) dt = dtfunc

!------------------------------------------------------------------------------
         cum_Timestep = cum_Timestep + 1
         sTime(1) = sTime(1) + dt
         sStep(1) = cum_Timestep

         ! Move the old timesteps one step down the ladder
         IF(timestep > 1 .OR. interval > 1) THEN
           DO i = SIZE(sPrevSizes,2),2,-1
             sPrevSizes(1,i) = sPrevSizes(1,i-1)
           END DO
           sPrevSizes(1,1) = sSize(1)
         END IF 
         sSize(1) = dt

         sInterval(1) = interval
         IF (.NOT. Transient ) steadyIt(1) = steadyIt(1) + 1
!------------------------------------------------------------------------------
         IF ( ParEnv % MyPE == 0 ) THEN
           CALL Info( 'MAIN', ' ', Level=3 )
           CALL Info( 'MAIN', '-------------------------------------', Level=3 )

           IF ( Transient .OR. Scanning ) THEN
             WRITE( Message, * ) 'Time: ',TRIM(i2s(cum_Timestep)),'/', &
                   TRIM(i2s(stepcount)), sTime(1)
             CALL Info( 'MAIN', Message, Level=3 )

             newtime= RealTime()

             IF( cum_Timestep > 1 ) THEN
               maxtime = ListGetConstReal( CurrentModel % Simulation,'Real Time Max',GotIt)
               IF( GotIt ) THEN
                  WRITE( Message,'(A,F8.3)') 'Fraction of real time left: ',&
                              1.0_dp-RealTime() / maxtime
               ELSE             
                 timeleft = NINT((stepcount-(cum_Timestep-1))*(newtime-prevtime)/60._dp);
                 IF (timeleft > 120) THEN
                   WRITE( Message, *) 'Estimated time left: ', &
                     TRIM(i2s(timeleft/60)),' hours.'
                 ELSE IF(timeleft > 60) THEN
                   WRITE( Message, *) 'Estimated time left: 1 hour ', &
                     TRIM(i2s(MOD(timeleft,60))), ' minutes.'
                 ELSE IF(timeleft >= 1) THEN
                   WRITE( Message, *) 'Estimated time left: ', &
                     TRIM(i2s(timeleft)),' minutes.'
                 ELSE
                   WRITE( Message, *) 'Estimated time left: less than a minute.'
                 END IF
               END IF
               CALL Info( 'MAIN', Message, Level=3 )
             END IF
             prevtime = newtime
           ELSE
             WRITE( Message, * ) 'Steady state iteration: ',cum_Timestep
             CALL Info( 'MAIN', Message, Level=3 )
           END IF

           CALL Info( 'MAIN', '-------------------------------------', Level=3 )
           CALL Info( 'MAIN', ' ', Level=3 )
         END IF

!------------------------------------------------------------------------------
!        Solve any and all governing equations in the system
!------------------------------------------------------------------------------
         AdaptiveTime = ListGetLogical( CurrentModel % Simulation, &
                  'Adaptive Timestepping', GotIt )

         IF ( Transient .AND. AdaptiveTime ) THEN 
            AdaptiveLimit = ListGetConstReal( CurrentModel % Simulation, &
                        'Adaptive Time Error', GotIt )
 
            IF ( .NOT. GotIt ) THEN 
               WRITE( Message, * ) 'Adaptive Time Limit must be given for' // &
                        'adaptive stepping scheme.'
               CALL Fatal( 'ElmerSolver', Message )
            END IF

            AdaptiveMaxTimestep = ListGetConstReal( CurrentModel % Simulation, &
                     'Adaptive Max Timestep', GotIt )
            IF ( .NOT. GotIt ) AdaptiveMaxTimestep =  dt
            AdaptiveMaxTimestep =  MIN(AdaptiveMaxTimeStep, dt)

            AdaptiveMinTimestep = ListGetConstReal( CurrentModel % Simulation, &
                     'Adaptive Min Timestep', GotIt )

            AdaptiveKeepSmallest = ListGetInteger( CurrentModel % Simulation, &
                       'Adaptive Keep Smallest', GotIt, minv=0  )

            n = CurrentModel % NumberOfSolvers
            j = 0
            k = 0
            DO i=1,n
               Solver => CurrentModel % Solvers(i)
               IF ( ASSOCIATED( Solver % Variable  % Values ) ) THEN
                  IF ( ASSOCIATED( Solver % Variable % PrevValues ) ) THEN
                     j = MAX( j, SIZE( Solver % Variable % PrevValues,2 ) )
                  END IF
                  k = MAX( k, SIZE( Solver % Variable % Values ) )
               END IF
            END DO
            ALLOCATE( xx(n,k), yynrm(n), xxnrm(n), prevxx( n,k,j ) )

            CumTime = 0.0d0
            IF ( ddt == 0.0d0 .OR. ddt > AdaptiveMaxTimestep ) ddt = AdaptiveMaxTimestep

            s = sTime(1) - dt
            SmallestCount = 0
            DO WHILE( CumTime < dt-1.0d-12 )
               ddt = MIN( dt - CumTime, ddt )

               DO i=1,CurrentModel % NumberOFSolvers
                  Solver => CurrentModel % Solvers(i)
                  IF ( ASSOCIATED( Solver % Variable % Values ) ) THEN
                     n = SIZE( Solver % Variable % Values )
                     xx(i,1:n) = Solver % Variable % Values
                     xxnrm(i) = Solver % Variable % Norm
                     IF ( ASSOCIATED( Solver % Variable % PrevValues ) ) THEN
                        DO j=1,SIZE( Solver % Variable % PrevValues,2 )
                           prevxx(i,1:n,j) = Solver % Variable % PrevValues(:,j)
                        END DO
                     END IF
                  END IF
               END DO

               sTime(1) = s + CumTime + ddt
               sSize(1) = ddt
               CALL SolveEquations( CurrentModel, ddt, Transient, &
                 CoupledMinIter, CoupledMaxIter, SteadyStateReached, RealTimestep )


               MaxErr = ListGetConstReal( CurrentModel % Simulation, &
                          'Adaptive Error Measure', GotIt )

               DO i=1,CurrentModel % NumberOFSolvers
                  Solver => CurrentModel % Solvers(i)
                  IF ( ASSOCIATED( Solver % Variable % Values ) ) THEN
                     n = SIZE(Solver % Variable % Values)
                     yynrm(i) = Solver % Variable % Norm
                     Solver % Variable % Values = xx(i,1:n)
                     IF ( ASSOCIATED( Solver % Variable % PrevValues ) ) THEN
                        DO j=1,SIZE( Solver % Variable % PrevValues,2 )
                           Solver % Variable % PrevValues(:,j) = prevxx(i,1:n,j)
                        END DO
                     END IF
                  END IF
               END DO

               sStep(1) = ddt / 2
               sTime(1) = s + CumTime + ddt/2
               CALL SolveEquations( CurrentModel, ddt/2, Transient, &
                  CoupledMinIter, CoupledMaxIter, SteadyStateReached, RealTimestep )
               sTime(1) = s + CumTime + ddt
               CALL SolveEquations( CurrentModel, ddt/2, Transient, &
                  CoupledMinIter, CoupledMaxIter, SteadyStateReached, RealTimestep )

               MaxErr = ABS( MaxErr - ListGetConstReal( CurrentModel % Simulation, &
                           'Adaptive Error Measure', GotIt ) )

               IF ( .NOT. GotIt ) THEN
                  MaxErr = 0.0d0
                  DO i=1,CurrentModel % NumberOFSolvers
                     Solver => CurrentModel % Solvers(i)
                     IF ( ASSOCIATED( Solver % Variable % Values ) ) THEN
                        IF ( yynrm(i) /= Solver % Variable % Norm ) THEN
                           Maxerr = MAX(Maxerr,ABS(yynrm(i)-Solver % Variable % Norm)/yynrm(i))
                        END IF
                     END IF
                  END DO
               END IF

               IF ( MaxErr < AdaptiveLimit .OR. ddt <= AdaptiveMinTimestep ) THEN
                 CumTime = CumTime + ddt
                 RealTimestep = RealTimestep+1
                 IF ( SmallestCount >= AdaptiveKeepSmallest .OR. StepControl > 0 ) THEN
                    ddt = MIN( 2*ddt, AdaptiveMaxTimeStep )
                    StepControl   = 1
                    SmallestCount = 0
                  ELSE
                    StepControl   = 0
                    SmallestCount = SmallestCount + 1
                  END IF
               ELSE
                  DO i=1,CurrentModel % NumberOFSolvers
                     Solver => CurrentModel % Solvers(i)
                     IF ( ASSOCIATED( Solver % Variable % Values ) ) THEN
                        n = SIZE(Solver % Variable % Values)
                        Solver % Variable % Norm = xxnrm(i)
                        Solver % Variable % Values = xx(i,1:n)
                        IF ( ASSOCIATED( Solver % Variable % PrevValues ) ) THEN
                           DO j=1,SIZE( Solver % Variable % PrevValues,2 )
                              Solver % Variable % PrevValues(:,j) = prevxx(i,1:n,j)
                           END DO
                        END IF
                     END IF
                  END DO
                  ddt = ddt / 2
                  StepControl = -1
               END IF
               WRITE(*,'(a,3e20.12)') 'Adaptive(cum,ddt,err): ', cumtime, ddt, maxerr
            END DO
            sSize(1) = dt
            sTime(1) = s + dt
  
            DEALLOCATE( xx, xxnrm, yynrm, prevxx )
         ELSE ! Adaptive timestepping
            CALL SolveEquations( CurrentModel, dt, Transient, &
              CoupledMinIter, CoupledMaxIter, SteadyStateReached, RealTimestep )
            RealTimestep = RealTimestep+1
         END IF
!------------------------------------------------------------------------------
!        Save results to disk, if requested
!------------------------------------------------------------------------------

         LastSaved = .FALSE.
         IF( OutputIntervals(Interval) /= 0 ) THEN

           CALL SaveToPost(0)
           k = MOD( Timestep-1, OutputIntervals(Interval) )
           IF ( k == 0 .OR. SteadyStateReached ) THEN
            
             DO i=1,CurrentModel % NumberOfSolvers
               Solver => CurrentModel % Solvers(i)
               IF ( Solver % Procedure == 0 ) CYCLE
               ExecThis = ( Solver % SolverExecWhen == SOLVER_EXEC_AHEAD_SAVE)
               When = ListGetString( Solver % Values, 'Exec Solver', GotIt )
               IF ( GotIt ) ExecThis = ( When == 'before saving') 
               IF( ExecThis ) CALL SolverActivate( CurrentModel,Solver,dt,Transient )
             END DO 

             CALL SaveCurrent(Timestep)
             LastSaved = .TRUE.

             DO i=1,CurrentModel % NumberOfSolvers
               Solver => CurrentModel % Solvers(i)
               IF ( Solver % Procedure == 0 ) CYCLE
               ExecThis = ( Solver % SolverExecWhen == SOLVER_EXEC_AFTER_SAVE)
               When = ListGetString( Solver % Values, 'Exec Solver', GotIt )
               IF ( GotIt ) ExecThis = ( When == 'after saving') 
               IF( ExecThis ) CALL SolverActivate( CurrentModel,Solver,dt,Transient )
             END DO 
           END IF
         END IF
!------------------------------------------------------------------------------

         maxtime = ListGetConstReal( CurrentModel % Simulation,'Real Time Max',GotIt)
         IF( GotIt .AND. RealTime() > maxtime ) THEN
            CALL Info('MAIN','Reached allowed maximum real time, exiting...')
            GOTO 100
         END IF

!------------------------------------------------------------------------------

         IF ( SteadyStateReached .AND. .NOT. (Transient .OR. Scanning) ) THEN
            IF ( Timestep >= CoupledMinIter ) EXIT
         END IF

!------------------------------------------------------------------------------
       END DO ! timestep within an iterval
!------------------------------------------------------------------------------
 
!------------------------------------------------------------------------------
     END DO ! timestep intervals, i.e. the simulation
!------------------------------------------------------------------------------

100   DO i=1,CurrentModel % NumberOfSolvers
        Solver => CurrentModel % Solvers(i)
        IF ( Solver % Procedure == 0 ) CYCLE
        When = ListGetString( Solver % Values, 'Exec Solver', GotIt )
        IF ( GotIt ) THEN
           IF ( When == 'after simulation' .OR. When == 'after all' ) THEN
              CALL SolverActivate( CurrentModel,Solver,dt,Transient )
              LastSaved = .FALSE.
           END IF
        ELSE
           IF ( Solver % SolverExecWhen == SOLVER_EXEC_AFTER_ALL ) THEN
              CALL SolverActivate( CurrentModel,Solver,dt,Transient )
              LastSaved = .FALSE.
           END IF
        END IF
     END DO

     IF( .NOT. LastSaved ) THEN
        DO i=1,CurrentModel % NumberOfSolvers
           Solver => CurrentModel % Solvers(i)
           IF ( Solver % Procedure == 0 ) CYCLE
           ExecThis = ( Solver % SolverExecWhen == SOLVER_EXEC_AHEAD_SAVE)
           When = ListGetString( Solver % Values, 'Exec Solver', GotIt )
           IF ( GotIt ) ExecThis = ( When == 'before saving') 
           IF( ExecThis ) CALL SolverActivate( CurrentModel,Solver,dt,Transient )
        END DO
     END IF

!------------------------------------------------------------------------------
   END SUBROUTINE ExecSimulation
!------------------------------------------------------------------------------


!------------------------------------------------------------------------------
  SUBROUTINE SaveCurrent( CurrentStep )
!------------------------------------------------------------------------------
    INTEGER :: i, CurrentStep,nlen
    TYPE(Variable_t), POINTER :: Var
    LOGICAL :: EigAnal, GotIt
    CHARACTER(LEN=MAX_NAME_LEN) :: Simul
    LOGICAL :: BinaryOutput, SaveAll
    
    Simul = ListGetString( CurrentModel % Simulation, 'Simulation Type' )
    
    OutputFile = ListGetString(CurrentModel % Simulation,'Output File',GotIt)
    IF ( GotIt ) THEN
      IF ( ParEnv % PEs > 1 ) THEN
        DO i=1,MAX_NAME_LEN
          IF ( OutputFile(i:i) == ' ' ) EXIT
        END DO
        OutputFile(i:i) = '.'
        WRITE( OutputFile(i+1:), '(a)' ) TRIM(i2s(ParEnv % MyPE))
      END IF

      BinaryOutput = ListGetLogical( CurrentModel % Simulation,'Binary Output',GotIt )
      IF ( .NOT.GotIt ) BinaryOutput = .FALSE.

      SaveAll = .NOT.ListGetLogical( CurrentModel % Simulation,&
                                    'Omit unchanged variables in output',GotIt )
      IF ( .NOT.GotIt ) SaveAll = .TRUE.

      Mesh => CurrentModel % Meshes
      DO WHILE( ASSOCIATED( Mesh ) ) 
         IF ( Mesh % OutputActive ) THEN
            nlen = LEN_TRIM(Mesh % Name )
            IF ( nlen > 0 ) THEN
               OutputName = Mesh % Name(1:nlen) // '/' // TRIM(OutputFile)
            ELSE
               OutputName = OutputFile
            END IF

            EigAnal = .FALSE.
            DO i=1,CurrentModel % NumberOfSolvers
               IF ( ASSOCIATED( CurrentModel % Solvers(i) % Mesh, Mesh ) ) THEN
                  EigAnal = ListGetLogical( CurrentModel % Solvers(i) % Values, &
                              'Eigen Analysis', GotIt )
                  EigAnal = EigAnal .OR. ListGetLogical( CurrentModel % Solvers(i) % Values, &
                              'Harmonic Analysis', GotIt )

                  IF ( EigAnal ) THEN
                     Var => CurrentModel % Solvers(i) % Variable
                     IF ( ASSOCIATED(Var % EigenValues) ) THEN
                        IF ( TotalTimesteps == 1 ) THEN
                           DO j=1,CurrentModel % Solvers(i) % NOFEigenValues
                             IF ( CurrentModel % Solvers(i) % Matrix % Complex ) THEN
                                DO k=1,SIZE(Var % Values)/2
                                   Var % Values(2*k-1) = REAL( Var % EigenVectors(j,k) )
                                   Var % Values(2*k-0) = AIMAG( Var % EigenVectors(j,k) )
                                 END DO
                              ELSE
                                 Var % Values = REAL( Var % EigenVectors(j,:) )
                              END IF
                              SavedSteps = SaveResult( OutputName, Mesh, &
                                     j, sTime(1), BinaryOutput, SaveAll )
                           END DO
                        ELSE
                           j = MIN( CurrentStep, SIZE( Var % EigenVectors,1 ) )
                           IF ( CurrentModel % Solvers(i) % Matrix % Complex ) THEN
                             DO k=1,SIZE(Var % Values)/2
                                Var % Values(2*k-1) = REAL( Var % EigenVectors(j,k) )
                                Var % Values(2*k-0) = AIMAG( Var % EigenVectors(j,k) )
                              END DO
                           ELSE
                              Var % Values = REAL(Var % EigenVectors(j,:))
                           END IF
                           SavedSteps = SaveResult( OutputName, Mesh, &
                                CurrentStep, sTime(1), BinaryOutput, SaveAll )
                        END IF
                        Var % Values = 0.0d0
                     END IF
                  END IF
               END IF
            END DO

            IF ( .NOT. EigAnal ) THEN
               SavedSteps = SaveResult( OutputName,Mesh, NINT(sStep(1)), &
                     sTime(1), BinaryOutput, SaveAll )
            END IF
         END IF
         Mesh => Mesh % Next
      END DO
    ELSE
      Mesh => CurrentModel % Meshes
      DO WHILE( ASSOCIATED( Mesh ) ) 
         IF ( Mesh % OutputActive ) &
           Mesh % SavesDone=Mesh % SavesDone+1
         Mesh => Mesh % Next
      END DO
    END IF
    CALL SaveToPost(CurrentStep)
!------------------------------------------------------------------------------
  END SUBROUTINE SaveCurrent
!------------------------------------------------------------------------------

!------------------------------------------------------------------------------
  SUBROUTINE SaveToPost(CurrentStep)
!------------------------------------------------------------------------------
!    Convert results file to post processing file, if requested
!------------------------------------------------------------------------------
    TYPE(Variable_t), POINTER :: Var
    LOGICAL :: EigAnal = .FALSE., Found
    INTEGER :: i, CurrentStep,nlen,timesteps
    CHARACTER(LEN=MAX_NAME_LEN) :: Simul, SaveWhich
    
    Simul = ListGetString( CurrentModel % Simulation,  'Simulation Type' )

    OutputFile = ListGetString( CurrentModel % Simulation,'Output File',GotIt )
    IF ( Gotit ) THEN
      IF ( ParEnv % PEs > 1 ) THEN
        DO i=1,MAX_NAME_LEN
          IF ( OutputFile(i:i) == ' ' ) EXIT
        END DO
        OutputFile(i:i) = '.'
        WRITE( OutputFile(i+1:), '(a)' ) TRIM(i2s(ParEnv % MyPE))
      END IF
    END IF
    
    PostFile = ListGetString( CurrentModel % Simulation,'Post File',GotIt )
    IF ( GotIt ) THEN
       IF ( ParEnv % PEs > 1 ) THEN
          DO i=1,MAX_NAME_LEN
             IF ( PostFile(i:i) == ' ' ) EXIT
          END DO
          PostFile(i:i) = '.'
          WRITE( PostFile(i+1:), '(a)' ) TRIM(i2s(ParEnv % MyPE))
       END IF
       
       Mesh => CurrentModel % Meshes
       DO WHILE( ASSOCIATED( Mesh ) )
          IF ( CurrentStep==0 .AND. Mesh % SavesDone>0) THEN
            Mesh => Mesh % Next; CYCLE
          END IF

          IF ( Mesh % OutputActive ) THEN
             nlen = LEN_TRIM(Mesh % Name)
             IF ( nlen==0 .OR. FileNameQualified(OutputFile) ) THEN
                OutputName = OutputFile
             ELSE
                OutputName = Mesh % Name(1:nlen)//'/'//TRIM(OutputFile)
             END IF

             IF ( nlen==0 .OR. FileNameQualified(PostFile) ) THEN
                PostName = PostFile
             ELSE
                Postname = Mesh % Name(1:nlen)//'/'//TRIM(PostFile)
             END IF

            IF ( ListGetLogical( CurrentModel % Simulation,'Filename Numbering',&
             GotIt) ) THEN
                IF( CurrentStep == 0 ) THEN
                  PostName = NextFreeFilename(PostName)
                ELSE 
                  PostName = NextFreeFilename(PostName,LastExisting = .TRUE. ) 
                END IF
             END IF

             CALL SetCurrentMesh( CurrentModel, Mesh )

             EigAnal = .FALSE.
             timesteps = TotalTimeSteps
             DO i=1,CurrentModel % NumberOfSolvers
                IF (ASSOCIATED(CurrentModel % Solvers(i) % Mesh, Mesh)) THEN
                   EigAnal = ListGetLogical( CurrentModel % &
                       Solvers(i) % Values, 'Eigen Analysis', GotIt )

                   EigAnal = EigAnal .OR. ListGetLogical( CurrentModel % &
                       Solvers(i) % Values, 'Harmonic Analysis', GotIt )

                   IF ( EigAnal ) timesteps = MAX( timesteps, &
                            CurrentModel % Solvers(i) % NOFEigenValues )
                END IF
             END DO

             DO i=1,CurrentModel % NumberOfSolvers
                IF (ASSOCIATED(CurrentModel % Solvers(i) % Mesh, Mesh)) THEN
                   EigAnal = ListGetLogical( CurrentModel % &
                       Solvers(i) % Values, 'Eigen Analysis', GotIt )

                   EigAnal = EigAnal .OR. ListGetLogical( CurrentModel % &
                       Solvers(i) % Values, 'Harmonic Analysis', GotIt )

                   IF ( EigAnal ) THEN
                      SaveWhich = ListGetString( CurrentModel % Solvers(i) % Values, &
                               'Eigen and Harmonic Solution Output', Found )

                      IF ( TotalTimesteps == 1 ) THEN
                         DO j=1,CurrentModel % Solvers(i) % NOFEigenValues
                           Var => Mesh % Variables
                           DO WHILE(ASSOCIATED(Var))
                             IF ( .NOT. ASSOCIATED(Var % EigenValues) ) THEN
                               Var => Var % Next
                               CYCLE
                             END IF
                             IF ( CurrentModel % Solvers(i) % Matrix % Complex ) THEN
                               DO k=1,SIZE(Var % Values)/2
                                 Var % Values(2*k-1) = REAL( Var % EigenVectors(j,k))
                                 Var % Values(2*k-0) = AIMAG(Var % EigenVectors(j,k))
                               END DO
                             ELSE
                                SELECT CASE(SaveWhich)
                                CASE('real part')
                                  Var % Values = Var % EigenVectors(j,:)
                                CASE('imag part')
                                  Var % Values = AIMAG(Var % EigenVectors(j,:))
                                CASE('abs value')
                                  Var % Values = ABS(Var % EigenVectors(j,:))
                                CASE('phase angle')
                                  Var % Values = ATAN2(AIMAG(Var % EigenVectors(j,:)), &
                                        REAL(Var % EigenVectors(j,:)))
                                CASE DEFAULT
                                  Var % CValues => Var % EigenVectors(j,:)
                                END SELECT
                             END IF
                             Var => Var % Next
                           END DO

                           IF ( CurrentStep > 0 ) THEN
                             IF ( Mesh % SavesDone /= 0 ) Mesh % SavesDone = j
                             CALL WritePostFile( PostName,OutputName, CurrentModel, &
                                CurrentModel % Solvers(i) % NOFEigenValues, .TRUE. )
                           END IF
                         END DO
                      ELSE
                         DO WHILE(ASSOCIATED(Var))
                           IF ( .NOT. ASSOCIATED(Var % EigenValues) ) THEN
                             Var => Var % Next
                             CYCLE
                           END IF
                           j = MIN( CurrentStep, SIZE( Var % EigenVectors,1 ) )
                           IF ( CurrentModel % Solvers(i) % Matrix % Complex ) THEN
                              DO k=1,SIZE(Var % Values)/2
                                 Var % Values(2*k-1) = REAL( Var % EigenVectors(j,k) )
                                 Var % Values(2*k-0) = AIMAG( Var % EigenVectors(j,k) )
                              END DO
                           ELSE
                              SELECT CASE(SaveWhich)
                              CASE('real part')
                                Var % Values = Var % EigenVectors(j,:)
                              CASE('imag part')
                                Var % Values = AIMAG(Var % EigenVectors(j,:))
                              CASE('abs value')
                                Var % Values = ABS(Var % EigenVectors(j,:))
                              CASE('phase angle')
                                Var % Values = ATAN2(AIMAG(Var % EigenVectors(j,:)), &
                                      REAL(Var % EigenVectors(j,:)))
                              CASE DEFAULT
                                Var % CValues => Var % EigenVectors(j,:)
                              END SELECT
                           END IF
                           Var => Var % Next
                         END DO

                         IF ( CurrentStep>0 ) THEN
                           IF ( Mesh % SavesDone /= 0 ) Mesh % SavesDone = CurrentStep
                           CALL WritePostFile( PostName,OutputName, CurrentModel, &
                             CurrentModel % Solvers(i) % NOFEigenValues, .TRUE. )
                         END IF
                      END IF
                      EXIT
                   END IF
                END IF
             END DO

             IF ( .NOT. EigAnal .OR. CurrentStep == 0 ) &
               CALL WritePostFile( PostName, OutputName, CurrentModel, timesteps, .TRUE. )

             Var => Mesh % Variables
             DO WHILE(ASSOCIATED(Var))
               IF ( ASSOCIATED(Var % EigenValues) ) THEN
                  Var % Values=0._dp
                  Var % CValues => NULL()
               END IF
               Var => Var % Next
             END DO
          END IF
          Mesh => Mesh % Next
       END DO
    END IF
!------------------------------------------------------------------------------
  END SUBROUTINE SaveToPost
!------------------------------------------------------------------------------


!------------------------------------------------------------------------------
  RECURSIVE SUBROUTINE FreeMesh(Mesh)
!------------------------------------------------------------------------------
    TYPE(Mesh_t), POINTER :: Mesh
!------------------------------------------------------------------------------
    IF (.NOT.ASSOCIATED(Mesh)) RETURN

    CALL FreeMesh(Mesh % Next)

    Mesh % Next   => NULL()
    Mesh % Child  => NULL()
    Mesh % Parent => NULL()

    CALL ReleaseMesh(Mesh)
    DEALLOCATE(Mesh)
!------------------------------------------------------------------------------
  END SUBROUTINE FreeMesh
!------------------------------------------------------------------------------


!------------------------------------------------------------------------------
  SUBROUTINE FreeSolver(Solver)
!------------------------------------------------------------------------------
    TYPE(Solver_t) :: Solver
!------------------------------------------------------------------------------
    CALL FreeValueList(Solver % Values)
    CALL FreeMatrix(Solver % Matrix)
    IF (ALLOCATED(Solver % Def_Dofs)) DEALLOCATE(Solver % Def_Dofs)
    IF (ASSOCIATED(Solver % ActiveElements)) DEALLOCATE(Solver % ActiveElements)
!------------------------------------------------------------------------------
  END SUBROUTINE FreeSolver
!------------------------------------------------------------------------------


!------------------------------------------------------------------------------
  SUBROUTINE FreeValueList(List)
!------------------------------------------------------------------------------
    TYPE(ValueList_t), POINTER :: List
!------------------------------------------------------------------------------
    TYPE(ValueList_t), POINTER :: ptr
   
    ptr => List
    DO WHILE(ASSOCIATED(ptr))
      IF (ASSOCIATED(ptr % TValues)) DEALLOCATE(ptr % TValues)
      IF (ASSOCIATED(ptr % FValues)) DEALLOCATE(ptr % FValues)
      IF (ASSOCIATED(ptr % IValues)) DEALLOCATE(ptr % IValues)
      ptr => ptr % Next
    END DO 
!------------------------------------------------------------------------------
  END SUBROUTINE FreeValueList
!------------------------------------------------------------------------------


!------------------------------------------------------------------------------
 SUBROUTINE FreeModel(Model)
!------------------------------------------------------------------------------
   TYPE(Model_t), POINTER :: Model
!------------------------------------------------------------------------------
   TYPE(Matrix_t), POINTER :: A,B
   INTEGER :: i
   IF (.NOT.ASSOCIATED(Model)) RETURN

   CALL FreeMesh(Model % Meshes)

   CALL FreeValueList(Model % Constants)
   CALL FreeValueList(Model % Simulation)

   IF (ASSOCIATED(Model % BCs)) THEN
     DO i=1,Model % NumberOfBCs
       IF (ASSOCIATED(Model % BCs(i) % PMatrix)) THEN
         A => Model % BCs(i) % PMatrix
         DO WHILE( ASSOCIATED(A) )
           B => A % Child
           A % Child => NULL()
           A => B
         END DO
         CALL FreeMatrix(Model % BCs(i) % PMatrix)
       END IF
       CALL FreeValueList( Model % BCs(i) % Values)
     END DO
     DEALLOCATE(Model % BCs)
   END IF

   DO i=1,Model % NumberOfSolvers
     CALL FreeSolver(Model % Solvers(i))
   END DO
   DEALLOCATE(Model % Solvers)

   IF (ASSOCIATED(Model % ICs)) THEN
     DO i=1,Model % NumberOfICs
       CALL FreeValueList( Model % ICs(i) % Values)
     END DO
     DEALLOCATE(Model % ICs)
   END IF

   IF (ASSOCIATED(Model % Bodies)) THEN
     DO i=1,Model % NumberOfBodies
       CALL FreeValueList( Model % Bodies(i) % Values)
     END DO
     DEALLOCATE(Model % Bodies)
   END IF

   IF (ASSOCIATED(Model % Equations)) THEN
     DO i=1,Model % NumberOfEquations
       CALL FreeValueList( Model % Equations(i) % Values)
     END DO
     DEALLOCATE(Model % Equations)
   END IF

   IF (ASSOCIATED(Model % BodyForces)) THEN
     DO i=1,Model % NumberOfBodyForces
       CALL FreeValueList( Model % BodyForces(i) % Values)
     END DO
     DEALLOCATE(Model % BodyForces)
   END IF

   Model=>NULL()
!------------------------------------------------------------------------------
 END SUBROUTINE FreeModel
!------------------------------------------------------------------------------

!------------------------------------------------------------------------------
  END SUBROUTINE ElmerSolver
!------------------------------------------------------------------------------
